
	.INCLUDE inc\global.inc
	.INCLUDE inc\dirp00.inc
	.INCLUDE inc\dirp01.inc
	.INCLUDE inc\fat.inc	

.LIST on

_FATTMP:	.SECTION page0, ref_only, offset _fattmp
FATTMP_START	.EQU	$

hidsec		.LWORD		; hidden sectors
totsec		.LWORD		; partition size in sector's
maxsec		.LWORD		; device max sector's count
freecnt		.DW		; free cluster's count
mbrflag		.EQU	freecnt	; mbr partition flag (low 8 bits only)
nextfree	.DW		; next free cluster
maxcls		.DW		; last usable cluster+1
mbrtype		.EQU	maxcls	; mbr partition type (low 8 bits only)
fatsize		.DW		; FAT16 fat table size
clscnt		.DW		; cluster's count #
ncnt		.DW		; generic counter

ldrv		.DB		; temp.: logical drive
ldttmp		.DB		; temp.: LDT pointer
pdev		.DB		; temp.: phisycal drive

FATTMP_SIZ	.EQU	($ - FATTMP_START)
	.ENDS


;---------------------------------------------------------------------------
; code segment -- bank $F9
;---------------------------------------------------------------------------

	.CODEF9

.comment @
; getfreecnt - compute free cluster's count scanning fat table
; in:	X = logical drive
; out:	LDT of logical drive is updated
;	CF = 1 if error
;
; fat table is loaded in dma buffer # $100, devoted to cluster cache and
; limited to 64 sectors
;
getfreecnt:
	stx	ldrv		; logical drive
	ldy	ldtp,x		; LDT pointer
	sty	ldttmp		; save LDT pointer
	stz	tlp		; init dma buffer pointer
	stz	tlp+1
	lda	#DMABNK
	sta	tlp+2
	lda	!P0LDT+ldt_fg1,y
	and	#00000011B
	sta	pdrive		; phisycal drive
	ACC16
	lda	!P0LDT+ldt_fat1,y
	sta	lba		; start sector of fat 1 table
	stz	lba+2
	stz	freecnt		; count of free clusters
	stz	nxtfree		; free next cluster
	stz	clscnt		; cluster numbering
	lda	!P0LDT+ldt_max,y
	sta	maxcls		; max. cluster number + 1
	lda	!P0LDT+ldt_fsiz,y
	sta	fatsize		; store fat table size

?lp:	lda	fatsize
	cmp	#CLSMAXSIZ
	bcc	?less
	lda	#CLSMAXSIZ
?less:	sta	wseccnt		; sectors count #
	ACC08
	lda	pdrive		; phisycal device #
	pha
	pei	(lba+2)		; start sector 
	pei	(lba)
	lda	seccnt		; sector's count
	pha
	pea	#CLSBUF		; dma cliuster buffer
	DISKRD			; read dik sector
	bcs	??		; error (Y = error code)

	ldy	#0
	CPU16
	lda	wseccnt
	xba
	asl	a		; bytes count #
	sta	ncnt
	ldx	clscnt		; X = current cluster number
?lpf:	lda	[tlp],y		; scan fat table
	bne	?nxt		; not free
	inc	freecnt		; update free cluster's count
	lda	nxtfree		; already set?
	bne	?nxt		; yes
	stx	nxtfree		; set first free cluster
?nxt:	inx			; update cluster number
	iny
	iny
	cpx	maxcls		; end of table?
	bcs	?done		; yes
	cpy	ncnt		; end of partial table?
	bcc	?lpf		; no
	stx	cluster		; save current cluster number
	INDEX08
	lda	fatsize
	sec
	sbc	wseccnt
	beq	?done		; end
	sta	fatsize
	clc
	lda	lba
	adc	wseccnt
	bcc	?lp
	inc	lba+2
	bra	?lp
?done:	CPU08
	clc
?err:	ACC16
	ldy	ldttmp		; LDT pointer
	lda	freecnt
	sta	!P0LDT+ldt_free,y
	lda	nxtfree
	sta	!P0LDT+ldt_nxt,y
	ACC08
	rts
@

.comment @
; load & check disk mbr
mbrld:
	lda	pdev		; phisycal device #
	pha
	pea	#0		; load sector #0
	pea	#0
	lda	#1		; load 1 sector
	pha
	pea	#$0100		; dma buffer #256
	DISKRD			; read dik sector
	bcs	??		; error (Y = error code)
	
	lda	#$10		; set bit mbr loaded
	sta	tmp
	CPU16
	ldy	#$01FE		; signature pointer
	lda	[bufp],y
	sta	mbrsgn		; mbr signature
	ldy	#$01C2		; partition type pointer
	lda	[bufp],y
	sta	wtmp		; partition type
	ldy	#$01C6		; partition start pointer
	lda	[bufp],y
	sta	nlba		; partition start
	iny
	iny
	lda	[bufp],y
	sta	nlba+2
	ldy	#$01BE		; partition flag pointer
	lda	[bufp],y
	sta	mbrfg		; partition flag
	lda	mbrsgn
	cmp	#$AA55
	CPU08
	bne	?20		; invalid mbr signature
	lda	#$18		; add valid mbr signature flag
	sta	tmp
	lda	mbrfg		; must be $80 or 0
	beq	?vv
	bpl	?20		; first partition invalid	
?vv:	lda	#$1C		; add active partition flag
	sta	tmp
	lda	drive
	asl	a
	asl	a
	tax			; index for access to atambr var
	lda	wtmp		; partition type
	tay
	sta	atambr+3,x	; store partition type
	lda	nlba+2		; store partition start
	sta	atambr+2,x
	lda	nlba+1
	sta	atambr+1,x
	lda	nlba
	sta	atambr,x
	lda	nlba+3		; must be 0
	bne	?20
	cpy	#$06		; just FAT16 type
	beq	?10
	cpy	#$04		; FAT 16 for disk less than 32Mb
	beq	?10
	cpy	#$0E
	bne	?20
?10:	CPU16
	ldy	#$01CA		; partition sec's count pointer
	lda	[bufp],y
	sta	ataprt,x
	iny
	iny
	lda	[bufp],y
	sta	ataprt+2,x
	CPU08
	lda	#$1D		; add valid partition type
	sta	tmp
?20:	ldx	drive		; update device flag
	lda	atadev,x
	ora	tmp
	sta	atadev,x
	clc
?30:	rts

; load & check pbr
	lda	pdev		; phisycal device #
	pha
	pei	(lba+2)		; load pbr sector 
	pea	(lba)
	lda	#1		; load 1 sector
	pha
	pea	#$0100		; dma buffer #256
	DISKRD			; read dik sector
	bcs	??		; error (Y = error code)
	
	lda	pdev		; phisycal drive
	asl	a
	asl	a
	asl	a
	tax
	ACC16
	lda	!atambr,x	; partition start
	sta	hidsec		; this is also hidden sectors count
	lda	!atambr+2,x
	and	#$00FF
	sta	hidsec+2	
	lda	!ataprt,x	; total partition sector's
	sta	totsec
	lda	!ataprt+2,x
	sta	totsec+2
	ldx	ldtcur		; LDT ptr

; entry point to check usb #0 disk
chk16c:
	ACC16
	ldx	pbrp+1
	inx
	stx	pbrp+1
	lda	#MAXF16
	sta	maxcnt
	ldy	#$FE		; pointer to pbr signature
	lda	[pbrp],y
	cmp	#$AA55
	bne	?10		; signature error
	dex
	stx	pbrp+1
	ldy	#$0B		; pointer to bytes per sector
	lda	[pbrp],y
	cmp	#$0200		; must be always 512 bytes per sector
	bne	?10		; error
	ldy	#$1C		; pointer to hidden sector's count
	lda	[pbrp],y	; typical value: $3F
	cmp	hidsec		; must match
	bne	?10
	iny
	iny
	lda	[pbrp],y
	cmp	hidsec+2
	bne	?10
	cmp	#0
	bne	?10		; must be 0
	ldy	#$0E		; pointer to reserved sector's
	lda	[pbrp],y	; typical value: 1 (no mandatory)
	sta	rsvsec
	ldy	#$11		; pointer to root dir entry #
	lda	[pbrp],y
	beq	?10		; fat32?
	cmp	#$0200		; typical value: not mandatory...
	bne	?10		; ...but this will be
	sta	rootent
	lsr	a		; divide by 16...
	lsr	a
	lsr	a
	lsr	a		;...to get root dir size in sec's
	sta	rootsiz
	ldy	#$13		; pointer to tot. sec16
	lda	[pbrp],y
	beq	?06		; invalid: is valid field tot. sec32
	cmp	totsec
	bne	?10		; error
	lda	totsec+2	; must be 0
	bne	?10		; errore
	beq	?08		; ok
?06:	ldy	#$20		; pointer to tot. sec32
	lda	[pbrp],y
	cmp	totsec
	bne	?10		; error
	iny
	iny
	lda	[pbrp],y
	cmp	totsec+2
	bne	?10		; error
?08:	ldy	#$16		; pointer to fat size (FAT16)
	lda	[pbrp],y
	sta	fatsiz
	beq	?10		; hmmm... is fat32... not good
	cmp	#257		; fat16 table size limited to 256 sec. max
	bra	?12
?10:	sec
?12:	CPU08
	bcs	?18		; error
	ldy	#$15		; pointer to media byte
	lda	[pbrp],y	; must be $F8
	cmp	#$F8
	bne	?16
	ldy	#$10		; pointer to count of fat table's
	lda	[pbrp],y
	cmp	#2		; must be 2
	bne	?16
	ldy	#$0D		; pointer to sector(s) per cluster
	lda	[pbrp],y
	ldx	#CLSCNT-1
?14:	cmp	>CLSSIZE,x
	beq	?20
	dex
	bpl	?14
?16:	sec
?18:	rts
?20:	sta	clssiz
	dec	a
	sta	clsmod
	lda	>CLSSHIFT,x
	sta	clsshf
	clc
	rts

@


;---------------------------------------------------------------------------
; tables
;---------------------------------------------------------------------------

; address of pbr buffer of fixed disk's
PBRBUFL:
	;.DB	<_pbr0, <_pbr1, <_pbr2, <_pbr3
PBRBUFH:
	.DB	>_pbr0, >_pbr1, >_pbr2, >_pbr3
	
; sector's per cluster table
CLSSIZE:
	.DB	$01, $02, $04, $08, $10, $20, $40

CLSCNT	.EQU	$-CLSSIZE

; sector's per cluster shift table   	
CLSSHIFT:
	.DB	$00, $01, $02, $03, $04, $05, $06
